/////////////////////////////////////
//File: computeRoburstNormalForm.h
//The front-end to the robust
//normal form computations
/////////////////////////////////////

//Author: Tomas Johnson
/////////////////////////////////////

#ifndef __computeNormalForm_h__
#define __computeNormalForm_h__

#include "interval.hpp"
#include "dim2taylor.hpp"

#include "fieldsAndCoord.h"

using namespace std;
using namespace cxsc;
using namespace taylor;

////////////////////////////////////////////////////////////////////////
bool computeRobustNormalForm(//INPUT:
			     
			     //THE STABLE EIGENVALUE
			     interval lambda_s, 
			     
			     //THE UNSTABLE EIGENVALUE
			     interval lambda_u,
			     
			     //THE NONLINEAR FIELD IN THE 
			     //STABLE DIRECTION
			     dim2taylor (*F_s)(dim2taylor_vector), 
			     
			     //THE NONLINEAR FIELD IN THE 
			     //UNSTABLE DIRECTION
			     dim2taylor (*F_u)(dim2taylor_vector),

			     //THE ORDER OF F
			     int order,
			     
			     //THE NONLINEAR FIELD IN THE 
			     //STABLE DIRECTION (complexified)
			     cinterval (*F_s_cplx)(civector), 
			     
			     //THE NONLINEAR FIELD IN THE 
			     //UNSTABLE DIRECTION(complexified)
			     cinterval (*F_u_cplx)(civector),

			     //IS F A POLYNOMIAL 
			     bool isPolynomial,

			     //Radius of Convergence of F_s/u
			     //Put negative for entire functions
			     real F_radiusOfConvergence,

			     //OUTPUT:
			     
			     //THE DOMAIN WHERE KAPPA IS VALID
			     real &r_2 , 
			     
			     //THE MAXIMAL DOMAIN WHERE YOU ARE
			     //RECOMMENDED TO SWITCH TO THE
			     //NORMAL FORM
			     real &r_3,
			     
			     //THE CONSTANT K_0 (BOUNDS THE NORM OF PHI)
			     real &K_0,

			     //THE CONSTANT KAPPA
			     real &kappa
			     );
/////////////////////////////////////////////////////////////////////////////

#endif
