#include "cimath.hpp"
#include "civector.hpp"
#include "interval.hpp"
#include "dim2taylor.hpp"

#include "computeRobustNormalForm.h"

using namespace std;
using namespace cxsc;
//using namespace taylor;

cinterval POWER(const cinterval &x, int n) {
  return power(x, n);
}
////////////////////////////////////////////////////////

//Use POWER(x,n), for the power function, 
//set order=degree(F) 
//Define constants as 'real'

////////////////////////////////////////////////////////
//MODIFY THESE:
//use x[1] for the stable direction,
//and x[2] for the unstable direction.
////////////////////////////////////////////////////////

//8L, constants
//eps=1e-3, l=10 good, eps_phi=0.15, eps=0.95
// real a(-0.788583173611695 ), b(0.634463132413968), 
//  c(-0.614928108234213 ), d(-0.772953125103622 ), A(-0.634661754756060), B(-0.634661754756060);
// real eps(1e-3);
//eps=1e-6, l=10 good, eps_phi=0.1, eps=0.9
// real a(-0.716193345867223 ), b(0.716036392706401), 
// c(-0.697901921000015), d(-0.698062951545206), A(-0.716271892254450), B(-0.716428897036110);
// real eps(1e-6);
//eps=1e-9, l=10 good, eps_phi=0.1, eps=0.9
// real a(-0.716114952516010 ), b( 0.716114795562846 ), 
//   c(-0.697982359936835 ), d( -0.697982520967383 ), A(-0.716350320908982), B(-0.716350477913766);
// real eps(1e-9);
real eps;
interval a, b, c, d, A, B;
real a_00(438.4905);

//The nonlinear part of the field
//in the stable direction:

template<class S, class T>
S Fs(T x) {
  //8L:
  S res =
    A*(POWER( a*x[1]+b*x[2],3)+sqr(a*x[1]+b*x[2])
	+ eps*(( (real)438.4905-(real)25.2469*
		 ((a*x[1]+b*x[2])-(real)452.7899*sqr(a*x[1]+b*x[2])))
	       *(c*x[1]+d*x[2])-(real)741.0341*POWER(c*x[1]+d*x[2],3)/ (real)3));

  return res;
}

//The nonlinear part of the field
//in the Unstable direction:

template<class S, class T>
S Fu(T x) {
  //8L:
  S res =B*(POWER( a*x[1]+b*x[2],3)+sqr(a*x[1]+b*x[2])
	+ eps*(( (real)438.4905-(real)25.2469*
		 ((a*x[1]+b*x[2])-(real)452.7899*sqr(a*x[1]+b*x[2])))
		    *(c*x[1]+d*x[2])-(real)741.0341*POWER(c*x[1]+d*x[2],3)/ (real)3));
  return res;
}



/////////////////////////////////////////////////////////

int main(int argc, char * argv[]) {
  eps=atof(argv[1]);
  ///////////////////////////////////////////////////////
  //MODIFY THESE:
  ///////////////////////////////////////////////////////
  cout<< SetPrecision(17,16);

  //The stable eigenvalue:
  //8L:, eps=0.001
  //interval lambda_s(-0.779788523026499);
  //eps=1e-6
  //interval lambda_s(-0.974460213889506);
  //eps=1e-9
  //interval lambda_s(-0.974679215235671);

  //The unstable eigenvalue:
  //8L:, eps=0.001
  //interval lambda_u(1.218279023026499);
  //eps=1e-6
  //interval lambda_u(0.974898704389506);
  //eps=1e-9
  //interval lambda_u(0.974679653726171);

  //The stable eigenvalue:
  interval a00e(a_00*eps/(real)2.0);
  interval lambda_s(a00e-sqrt(sqr(a00e)+interval(19.0)/(real)20));
  //The unstable eigenvalue:
  interval lambda_u(a00e+sqrt(sqr(a00e)+interval(19.0)/(real)20));

  interval sfactor(lambda_u*interval(20)/(real)19);
  interval ufactor(lambda_s*interval(20)/(real)19);
  interval snorm(sqrt1px2(sfactor));
  interval unorm(sqrt1px2(ufactor));
  a=-sfactor/snorm;
  b=-ufactor/unorm;
  c=-interval(1.0)/snorm;
  d=-interval(1.0)/unorm;
  interval adMbc(a*d-b*c);
  A=-b/adMbc;
  B=a/adMbc;
  cout<<"The stable eigenvalue   = "<<lambda_s<<endl
      <<"The unstable eigenvalue = "<<lambda_u<<endl;
  //cout<<a<<" "<<b<<endl<<c<<" "<<d<<endl<<A<<endl<<B<<endl;
  //The degree of F
  int order(20);

  //Is F a polynomial of degree 
  //less than or equal to order
  bool isPolynomial=true;

  //If F_s/u are entire put radiusIfConvergence < 0 (=-1.0)
  real F_radiusOfConvergence(-1.0);

  ////////////////////////////////////////////////////////
  //DO NOT MODIFY BELOW !
  ////////////////////////////////////////////////////////
 

  dim2taylor (*F_s)(dim2taylor_vector);
  dim2taylor (*F_u)(dim2taylor_vector);
  F_s=Fs<dim2taylor, dim2taylor_vector>;
  F_u=Fu<dim2taylor, dim2taylor_vector>;

  cinterval (*F_s_cplx)(civector);
  cinterval (*F_u_cplx)(civector);
  F_s_cplx=Fs<cinterval,civector>;
  F_u_cplx=Fu<cinterval,civector>;

  real r_2, r_3, K_0, kappa;
  cout<< SetPrecision(3,2);
  computeRobustNormalForm(lambda_s, lambda_u, F_s, F_u,
			  order,  F_s_cplx, F_u_cplx,
			  isPolynomial, F_radiusOfConvergence,
			  r_2, r_3, K_0, kappa);
  cout<< SetPrecision(7,6);
  cout<<"RESULTS"<<endl
      <<"r_2   = "<<r_2<<endl
      <<"r_3   = "<<r_3<<endl
      <<"K_0   = "<<K_0<<endl
      <<"kappa = "<<kappa<<endl<<endl;
}
