#include "cimath.hpp"
#include "civector.hpp"
#include "interval.hpp"
#include "dim2taylor.hpp"

#include "computeRobustNormalForm.h"

using namespace std;
using namespace cxsc;
//using namespace taylor;

cinterval POWER(const cinterval &x, int n) {
  return power(x, n);
}
////////////////////////////////////////////////////////

//Use POWER(x,n), for the power function, 
//set order=degree(F) 
//Define constants as 'real'

////////////////////////////////////////////////////////
//MODIFY THESE:
//use x[1] for the stable direction,
//and x[2] for the unstable direction.
////////////////////////////////////////////////////////

//The nonlinear part of the field
//in the stable direction:

template<class S, class T>
S Fs(T x) {
  
  S res = -interval(0.1)*POWER(x[1],3)-sqr(x[1])*x[2]+interval(1.3)*sqr(x[1])+2*x[1]*x[2];
  
  return res;
}

//The nonlinear part of the field
//in the Unstable direction:

template<class S, class T>
S Fu(T x) {
  
  S res = -interval(0.1)*POWER(x[2],3)+sqr(x[2])*x[1]-interval(0.7)*sqr(x[2])-2*x[1]*x[2]; ;

  return res;
}



/////////////////////////////////////////////////////////

int main() {
  ///////////////////////////////////////////////////////
  //MODIFY THESE:
  ///////////////////////////////////////////////////////
  cout<< SetPrecision(5,4);
  //The stable eigenvalue:
  
  interval lambda_s(-2.2);
  //The unstable eigenvalue:
  interval lambda_u(1.8);

  //  cout<<"The stable eigenvalue   = "<<lambda_s<<endl
  //   <<"The unstable eigenvalue = "<<lambda_u<<endl;
  
  //The degree of F
  int order(3);

  //Is F a polynomial of degree 
  //less than or equal to order
  bool isPolynomial=true;

  //If F_s/u are entire put radiusIfConvergence < 0 (=-1.0)
  real F_radiusOfConvergence(-1.0);

  ////////////////////////////////////////////////////////
  //DO NOT MODIFY BELOW !
  ////////////////////////////////////////////////////////
 

  dim2taylor (*F_s)(dim2taylor_vector);
  dim2taylor (*F_u)(dim2taylor_vector);
  cinterval (*F_s_cplx)(civector);
  cinterval (*F_u_cplx)(civector); 

  F_s=Fs<dim2taylor, dim2taylor_vector>;
  F_u=Fu<dim2taylor, dim2taylor_vector>;
  F_s_cplx=Fs<cinterval,civector>;
  F_u_cplx=Fu<cinterval,civector>;  
  

  real r_2, r_3, K_0, kappa;

  computeRobustNormalForm(lambda_s, lambda_u, F_s, F_u,
			  order,  F_s_cplx, F_u_cplx,
			  isPolynomial, F_radiusOfConvergence,
			  r_2, r_3, K_0, kappa);
  cout<< SetPrecision(7,6);
  cout<<"RESULTS"<<endl
      <<"r_2   = "<<r_2<<endl
      <<"r_3   = "<<r_3<<endl
      <<"K_0   = "<<K_0<<endl
      <<"kappa = "<<kappa<<endl<<endl;
}
